package com.ywt.gateway.configuration;


import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.web.reactive.error.ErrorWebExceptionHandler;
import org.springframework.core.Ordered;
import org.springframework.core.annotation.Order;
import org.springframework.core.io.buffer.DataBufferFactory;
import org.springframework.http.HttpStatus;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.stereotype.Component;
import org.springframework.web.server.ResponseStatusException;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Mono;

/**
 * @company: kayakwise
 * @author: TWX
 * @date: 2023/4/18 15:55
 * @description: 网关全局异常处理器设置响应内容 避免上游xml格式，却返回json
 */
@Component
@Slf4j
@Order(Ordered.HIGHEST_PRECEDENCE)
public class GlobalExceptionConfiguration implements ErrorWebExceptionHandler {
    //日志对象
    @Override //参数1: request response   ex:出现异常时异常对象
    public Mono<Void> handle(ServerWebExchange exchange, Throwable ex) {
        log.error("网关出现异常:", ex);
        // 1.获取响应对象
        ServerHttpResponse response = exchange.getResponse();
        // 2.response是否结束  用于多个异常处理时候
        if (response.isCommitted()) {
            return Mono.error(ex);
        }
        // 3.设置响应头类型
        response.getHeaders().setContentType(exchange.getRequest().getHeaders().getContentType());
        // 4.设置响应码
        if (ex instanceof ResponseStatusException) {
            response.setStatusCode(((ResponseStatusException) ex).getStatus());
        } else {
            response.setStatusCode(HttpStatus.SERVICE_UNAVAILABLE);
        }
        // 5.设置响应内容 避免上游xml格式，却返回json
        return response.writeWith(Mono.fromSupplier(() -> {
                    DataBufferFactory bufferFactory = response.bufferFactory();
                    //设置响应到response的数据
                    return bufferFactory.wrap(new byte[]{});
                }));
    }
}
