package com.ywt.gateway.decorator;

import org.reactivestreams.Publisher;
import org.springframework.cloud.gateway.filter.factory.rewrite.GzipMessageBodyResolver;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.core.io.buffer.DataBufferFactory;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.core.io.buffer.DefaultDataBufferFactory;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.http.server.reactive.ServerHttpResponseDecorator;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.List;

/**
 * @author Walker
 * Created on 2023/11/9
 */
public class CountResponseBodyDecorator extends ServerHttpResponseDecorator {
    private byte[] body;
    GzipMessageBodyResolver gzipMessageBodyResolver = new GzipMessageBodyResolver();

    public CountResponseBodyDecorator(ServerHttpResponse delegate) {
        super(delegate);
    }

    @Override
    public Mono<Void> writeWith(Publisher<? extends DataBuffer> publisher) {

        MediaType mediaType = getDelegate().getHeaders().getContentType();
        boolean isApplicationJsonMedia = MediaType.APPLICATION_JSON.isCompatibleWith(mediaType);
        DataBufferFactory bufferFactory = getDelegate().bufferFactory();

        if (publisher instanceof Flux && isApplicationJsonMedia) {
            Flux<? extends DataBuffer> fluxBody = Flux.from(publisher);
            return super.writeWith(fluxBody.buffer().map(dataBuffer -> {
                DataBuffer join = null;
                try {
                    //解决如果响应过大，会进行截断，出现乱码的问题
                    DataBufferFactory dataBufferFactory = DefaultDataBufferFactory.sharedInstance;
                    join = dataBufferFactory.join(dataBuffer);
                    body = new byte[join.readableByteCount()];
                    join.read(body);

                    //判断响应头的encoding-type是否为gzip，如果是则进行解压缩处理
                    List<String> encodingList = getDelegate().getHeaders().get(HttpHeaders.CONTENT_ENCODING);
                    String zipEncodingType = gzipMessageBodyResolver.encodingType();
                    boolean zip = encodingList != null && encodingList.contains(zipEncodingType);

                    DataBuffer wrapDataBuffer = bufferFactory.wrap(body);

                    body = zip ? gzipMessageBodyResolver.decode(body) : body;

                    return wrapDataBuffer;
                } finally {
                    if (join != null) {
                        //释放掉内存
                        DataBufferUtils.release(join);
                    }
                }
            }));
        }
        return super.writeWith(publisher);
    }

    @Override
    public Mono<Void> writeAndFlushWith(Publisher<? extends Publisher<? extends DataBuffer>> body) {
        return writeWith(Flux.from(body).flatMapSequential(p -> p));
    }
}
