package com.ywt.gateway.filter;

import com.ywt.gateway.model.CountRecord;
import com.ywt.gateway.model.RequestRecord;
import lombok.extern.slf4j.Slf4j;
import org.reactivestreams.Publisher;
import org.redisson.api.RDeque;
import org.redisson.api.RedissonClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cloud.gateway.filter.GatewayFilterChain;
import org.springframework.cloud.gateway.filter.GlobalFilter;
import org.springframework.cloud.gateway.filter.factory.rewrite.CachedBodyOutputMessage;
import org.springframework.cloud.gateway.support.BodyInserterContext;
import org.springframework.core.Ordered;
import org.springframework.core.ResolvableType;
import org.springframework.core.io.buffer.*;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.codec.CodecConfigurer;
import org.springframework.http.codec.HttpMessageReader;
import org.springframework.http.codec.multipart.FormFieldPart;
import org.springframework.http.codec.multipart.Part;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.http.server.reactive.ServerHttpRequestDecorator;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.http.server.reactive.ServerHttpResponseDecorator;
import org.springframework.stereotype.Component;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.reactive.function.BodyInserter;
import org.springframework.web.reactive.function.BodyInserters;
import org.springframework.web.reactive.function.server.HandlerStrategies;
import org.springframework.web.reactive.function.server.ServerRequest;
import org.springframework.web.server.ServerWebExchange;
import org.springframework.web.util.UriComponentsBuilder;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.nio.charset.StandardCharsets;
import java.util.*;

/**
 * 统计&日志拦截器
 *
 * @author Walker
 * Created on 2023/11/8
 *
 */
@Slf4j
@Component
public class GlobalCountAndLogFilter implements GlobalFilter, Ordered {
    @Autowired
    private CodecConfigurer codecConfigurer;
//    private final List<HttpMessageReader<?>> messageReaders = codecConfigurer.getReaders();
    private final Logger logger = LoggerFactory.getLogger(GlobalCountAndLogFilter.class);
    public static final String COUNTER_QUEUE_NAME = "gw:counter-queue";
    public static final String LOG_QUEUE_NAME = "gw:log-queue";

    @Autowired
    private RedissonClient redissonClient;

    @Override
    public int getOrder() {
        return -233;
    }

    @Override
    @SuppressWarnings("unchecked")
    public Mono<Void> filter(ServerWebExchange exchange, GatewayFilterChain chain) {
        ServerHttpRequest request = exchange.getRequest();
        CountRecord countRecord = new CountRecord();
        RequestRecord requestRecord = new RequestRecord();
        String requestUrl = request.getPath().toString();
        String methodName = request.getMethodValue();
        String requestId = request.getId();
        String backendUrl = request.getPath().pathWithinApplication().value();
        String ip = getIpAddress(request);
        long now = new Date().getTime();
        String host = "";
        if (request.getLocalAddress() != null) {
            host = request.getLocalAddress().getHostString();
        } else {
            logger.warn("Could not get local address!");
        }
        // TODO: 如何获取
        String backendAddr = "";
        String referer = "";
        String appId = "";
        String cliVer = "";
        String authData = "";
        String authParam = "";

        countRecord.setHost(host);
        countRecord.setRequestURI(requestUrl);
        countRecord.setMethod(methodName);
        countRecord.setRequestID(requestId);
        countRecord.setBackendURI(backendUrl);
        countRecord.setRequestOn(now);
        countRecord.setIp(ip);
        countRecord.setReferer(referer);
        countRecord.setAppid(appId);
        countRecord.setBackendAddr(backendAddr);

        requestRecord.setHost(host);
        requestRecord.setRequestURI(requestUrl);
        requestRecord.setMethod(methodName);
        requestRecord.setRequestID(requestId);
        requestRecord.setBackendURI(backendUrl);
        requestRecord.setRequestOn(now);
        requestRecord.setCliVer(cliVer);
        requestRecord.setAuthData(authData);
        requestRecord.setAuthParam(authParam);
        requestRecord.setBackendAddr(backendAddr);

        MediaType mediaType = request.getHeaders().getContentType();
        if (MediaType.APPLICATION_JSON.isCompatibleWith(mediaType)) {
            // JSON
            return writeBodyLog(exchange, chain, countRecord, requestRecord);
        } else if (MediaType.MULTIPART_FORM_DATA.isCompatibleWith(mediaType) || MediaType.APPLICATION_FORM_URLENCODED.isCompatibleWith(mediaType)) {
            // form-data
            return readFormData(exchange, chain, countRecord, requestRecord);
        } else {
            return writeBasicLog(exchange, chain, countRecord, requestRecord);
        }
    }

    private Mono<Void> writeBasicLog(ServerWebExchange exchange, GatewayFilterChain chain, CountRecord countRecord, RequestRecord requestRecord) {
        // 异步处理兼容处理get请求
        if("GET".equalsIgnoreCase(exchange.getRequest().getMethodValue())){
            StringBuilder builder = new StringBuilder();
            MultiValueMap<String, String> queryParams = exchange.getRequest().getQueryParams();
            if (!queryParams.isEmpty()) {
                for (Map.Entry<String, List<String>> entry : queryParams.entrySet()) {
                    builder.append(entry.getKey()).append("=").append(String.join(",", entry.getValue()));
                }
            }
            requestRecord.setRequestBody(builder.toString());
            //获取响应体进行包装
            ServerHttpResponseDecorator decoratedResponse = recordResponseLog(exchange, countRecord, requestRecord);

            // 打印日志
            commitRecord(countRecord, requestRecord);

            // 传递到下一个过滤器
            return chain.filter(exchange.mutate().response(decoratedResponse).build());
        }

        return DataBufferUtils.join(exchange.getRequest().getBody()).flatMap(dataBuffer -> {
            DataBufferUtils.retain(dataBuffer);
            final Flux<DataBuffer> cachedFlux = Flux.defer(() -> Flux.just(dataBuffer.slice(0, dataBuffer.readableByteCount())));
            final ServerHttpRequest mutatedRequest = new ServerHttpRequestDecorator(exchange.getRequest()) {
                @Override
                public Flux<DataBuffer> getBody() {
                    return cachedFlux;
                }
                @Override
                public MultiValueMap<String, String> getQueryParams() {
                    return UriComponentsBuilder.fromUri(exchange.getRequest().getURI()).build().getQueryParams();
                }
            };
            StringBuilder builder = new StringBuilder();
            MultiValueMap<String, String> queryParams = exchange.getRequest().getQueryParams();
            if (!queryParams.isEmpty()) {
                for (Map.Entry<String, List<String>> entry : queryParams.entrySet()) {
                    builder.append(entry.getKey()).append("=").append(String.join(",", entry.getValue()));
                }
            }
            requestRecord.setRequestBody(builder.toString());
            //获取响应体
            ServerHttpResponseDecorator decoratedResponse = recordResponseLog(exchange, countRecord, requestRecord);
            return chain.filter(exchange.mutate().request(mutatedRequest).response(decoratedResponse).build())
                    .then(Mono.fromRunnable(() -> {
                        // 打印日志
                        commitRecord(countRecord, requestRecord);
                    }));
        });
    }

    /**
     * 读取form-data数据
     */
    private Mono<Void> readFormData(ServerWebExchange exchange, GatewayFilterChain chain, CountRecord countRecord, RequestRecord requestRecord) {
        return DataBufferUtils.join(exchange.getRequest().getBody()).flatMap(dataBuffer -> {
            DataBufferUtils.retain(dataBuffer);
            final Flux<DataBuffer> cachedFlux = Flux.defer(() -> Flux.just(dataBuffer.slice(0, dataBuffer.readableByteCount())));
            final ServerHttpRequest mutatedRequest = new ServerHttpRequestDecorator(exchange.getRequest()) {
                @Override
                public Flux<DataBuffer> getBody() {
                    return cachedFlux;
                }
                @Override
                public MultiValueMap<String, String> getQueryParams() {
                    return UriComponentsBuilder.fromUri(exchange.getRequest().getURI()).build().getQueryParams();
                }
            };
            final HttpHeaders headers = exchange.getRequest().getHeaders();
            if (headers.getContentLength() == 0) {
                return chain.filter(exchange);
            }
            ResolvableType resolvableType;
            if (MediaType.MULTIPART_FORM_DATA.isCompatibleWith(headers.getContentType())) {
                resolvableType = ResolvableType.forClassWithGenerics(MultiValueMap.class, String.class, Part.class);
            } else {
                //解析 application/x-www-form-urlencoded
                resolvableType = ResolvableType.forClass(String.class);
            }
            return codecConfigurer.getReaders().stream().filter(reader -> reader.canRead(resolvableType, mutatedRequest.getHeaders().getContentType())).findFirst()
                    .orElseThrow(() -> new IllegalStateException("no suitable HttpMessageReader.")).readMono(resolvableType, mutatedRequest, Collections.emptyMap()).flatMap(resolvedBody -> {
                        if (resolvedBody instanceof MultiValueMap) {
                            LinkedMultiValueMap map = (LinkedMultiValueMap) resolvedBody;
                            if (!map.isEmpty()) {
                                StringBuilder builder = new StringBuilder();
                                // 字段解析
                                String key = "data";
                                final Part bodyPartInfo = (Part) ((MultiValueMap) resolvedBody).getFirst(key);
                                if (bodyPartInfo instanceof FormFieldPart) {
                                    String body = ((FormFieldPart) bodyPartInfo).value();
                                    builder.append(key).append("=").append(body);
                                }
                                requestRecord.setRequestBody(builder.toString());
                            }
                        } else {
                            requestRecord.setRequestBody((String) resolvedBody);
                        }
                        //获取响应体
                        ServerHttpResponseDecorator decoratedResponse = recordResponseLog(exchange, countRecord, requestRecord);
                        return chain.filter(exchange.mutate().request(mutatedRequest).response(decoratedResponse).build())
                                .then(Mono.fromRunnable(() -> {
                                    // 打印日志
                                    commitRecord(countRecord, requestRecord);
                                }));
                    });

        });
    }


    /**
     * 解决 request body 只能读取一次问题，
     * 参考: org.springframework.cloud.gateway.filter.factory.rewrite.ModifyRequestBodyGatewayFilterFactory
     */
    @SuppressWarnings("unchecked")
    private Mono writeBodyLog(ServerWebExchange exchange, GatewayFilterChain chain, CountRecord countRecord, RequestRecord requestRecord) {
        ServerRequest serverRequest = ServerRequest.create(exchange, codecConfigurer.getReaders());
        Mono<String> modifiedBody = serverRequest.bodyToMono(String.class)
                .flatMap(body -> {
                    requestRecord.setRequestBody(body);
                    return Mono.just(body);
                });
        // 通过 BodyInserter 插入 body 修改, 避免 request body 只能获取一次
        BodyInserter bodyInserter = BodyInserters.fromPublisher(modifiedBody, String.class);
        HttpHeaders headers = new HttpHeaders();
        headers.putAll(exchange.getRequest().getHeaders());
        // the new content type will be computed by bodyInserter
        // and then set in the request decorator
        headers.remove(HttpHeaders.CONTENT_LENGTH);
        CachedBodyOutputMessage outputMessage = new CachedBodyOutputMessage(exchange, headers);
        return bodyInserter.insert(outputMessage, new BodyInserterContext())
                .then(Mono.defer(() -> {
                    // 重新封装请求
                    ServerHttpRequest decoratedRequest = requestDecorate(exchange, headers, outputMessage);
                    // 记录响应日志
                    ServerHttpResponseDecorator decoratedResponse = recordResponseLog(exchange, countRecord, requestRecord);
                    return chain.filter(exchange.mutate().request(decoratedRequest).response(decoratedResponse).build())
                            .then(Mono.fromRunnable(() -> {
                                commitRecord(countRecord, requestRecord);
                            }));
                }));
    }

    /**
     * 记录日志
     */
    private void commitRecord(CountRecord countRecord, RequestRecord requestRecord) {
        RDeque<String> reqDeque = redissonClient.getDeque(LOG_QUEUE_NAME);
        reqDeque.addFirst(requestRecord.toString());
        RDeque<String> rDeque = redissonClient.getDeque(COUNTER_QUEUE_NAME);
        rDeque.addFirst(countRecord.toString());
    }

    /**
     * 请求装饰器，重新计算 headers
     */
    private ServerHttpRequestDecorator requestDecorate(ServerWebExchange exchange, HttpHeaders headers, CachedBodyOutputMessage outputMessage) {
        return new ServerHttpRequestDecorator(exchange.getRequest()) {
            @Override
            public HttpHeaders getHeaders() {
                long contentLength = headers.getContentLength();
                HttpHeaders httpHeaders = new HttpHeaders();
                httpHeaders.putAll(super.getHeaders());
                if (contentLength > 0) {
                    httpHeaders.setContentLength(contentLength);
                } else {
                    httpHeaders.set(HttpHeaders.TRANSFER_ENCODING, "chunked");
                }
                return httpHeaders;
            }
            @Override
            public Flux<DataBuffer> getBody() {
                return outputMessage.getBody();
            }
        };
    }

    /**
     * 记录响应日志
     * 通过 DataBufferFactory 解决响应体分段传输问题
     */
    private ServerHttpResponseDecorator recordResponseLog(ServerWebExchange exchange, CountRecord countRecord, RequestRecord requestRecord) {
        ServerHttpResponse response = exchange.getResponse();
        DataBufferFactory bufferFactory = response.bufferFactory();
        return new ServerHttpResponseDecorator(response) {
            @Override
            public Mono<Void> writeWith(Publisher<? extends DataBuffer> body) {
                if (body instanceof Flux) {
                    Date responseTime = new Date();
                    // 计算执行时间
                    long executeTime = (responseTime.getTime() - requestRecord.getRequestOn());
                    int statusCode = this.getStatusCode() == null ? 0 : this.getStatusCode().value();
                    String message = this.getStatusCode() == null ? "" : this.getStatusCode().getReasonPhrase();
                    requestRecord.setConsume((int) executeTime);
                    requestRecord.setStatusCode(statusCode);
                    countRecord.setConsume((int) executeTime);
                    countRecord.setStatusCode(statusCode);
                    countRecord.setMessage(message);

                    Flux<? extends DataBuffer> fluxBody = Flux.from(body);
                    return super.writeWith(fluxBody.buffer().map(dataBuffers -> {
                        // 合并多个流集合，解决返回体分段传输
                        DataBufferFactory dataBufferFactory = new DefaultDataBufferFactory();
                        DataBuffer join = dataBufferFactory.join(dataBuffers);
                        byte[] content = new byte[join.readableByteCount()];
                        join.read(content);
                        // 释放内存
                        DataBufferUtils.release(join);

                        String responseResult = new String(content, StandardCharsets.UTF_8);
                        requestRecord.setResponseBody(responseResult);
                        return bufferFactory.wrap(content);
                    }));
                }
                return super.writeWith(body);
            }
        };
    }

    /**
     * 获取 IP 地址
     */
    public static String getIpAddress(ServerHttpRequest request) {
        HttpHeaders headers = request.getHeaders();
        String ip = headers.getFirst("x-forwarded-for");
        if (ip != null && ip.length() != 0 && !"unknown".equalsIgnoreCase(ip)) {
            // 多次反向代理后会有多个ip值，第一个ip才是真实ip
            if (ip.contains(",")) {
                ip = ip.split(",")[0];
            }
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = headers.getFirst("Proxy-Client-IP");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = headers.getFirst("WL-Proxy-Client-IP");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = headers.getFirst("HTTP_CLIENT_IP");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = headers.getFirst("HTTP_X_FORWARDED_FOR");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = headers.getFirst("X-Real-IP");
        }
        if (ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)) {
            ip = request.getRemoteAddress() == null ? "" : request.getRemoteAddress().getAddress().getHostAddress();
        }
        return ip.equals("0:0:0:0:0:0:0:1") ? "127.0.0.1" : ip;
    }
}